#ifndef __DRIVER_H__
#define __DRIVER_H__

#include "window.h"
#include "haplotype.h"
#include "model.h"

class Driver
{
 public:

    Driver(CSample &s, CDatfile &d, CMap &m) : sample(s), dat(d), locmap(m) { }

    // Command-line argument list
    CArgs *a;

    // Marker positions from --alt and --null
    vector<int> alt;
    vector<int> null;

    // Extracted phenotype and haplotype data
    CSample & sample;
    CDatfile & dat;
    CMap & locmap;

    // Details of sliding windows
    vector<Window*> win;

    // Overall notal number of loci/snps in sample (for window)  
    int total_nloci;
    int total_nsnps;

    // User-specified constraints
    vector<int> usr_alt;
    vector<int> usr_null;
    
    // Number of haplotype-specific test
    int hs;

    // Storage for all pre-permuted phenotypes
    CMatrix preperm;
    vector<bool> original_bt;
    vector<double> original_qt;
    CMatrix wpreperm;

    // Case/control haplotype frequencies
    CVector hf_case;
    CVector hf_control;

    // Misc. : 'dummy' phases for denominator in coditional analysis
    vector<CPhase> poss;
    vector<CParentalPhase> pposs;
    CIndividual * everyman; 
    CIndividual * everytrio;

    // Microsatellite downcoding
    // e.g. SNP, SNP, usat(3), SNP, usat(4), SNP
    vector<int> acount;   //  1 1 3     1 4       1
    vector< vector<int> > downcode;  // [1] = 1
                                     // [2] = 2
                                     // [3] = 3,4,5
                                     // [4] = 6
                                     // [5] = 7,8,9,10
                                     // [6] = 11
       

    // Functions

    void firstPass();
    void initWindow();
    void slideWindow_BASIC();
    void slideWindow_CONSTANT_HAP();
    void initWhap();
    void processWindow();
    void prePermute();
    void prePermuteW();
    void storeOriginal();
    void permuteW(CMatrix&);

    void scoreWithoutParents(vector<CPhase>&,
			     CColVector&,
			     CMatrix&,
			     CMatrix&, 
			     int&, 
			     int,
			     CIndividual *,
			     int);
    
    void scoreWithParents(vector<CParentalPhase>&,
			  CColVector&,
			  CMatrix&,
			  CMatrix&, 
			  int&, 
			  int,
			  CIndividual *,
			  int);

    vector<CPhase> allPossibleWithoutParents(int, int);
    vector<CParentalPhase> allPossibleWithParents(int,int);
    
};


#endif
