#include "helper.h"
#include <iomanip>
#include <fstream>
#include <string>

StringArray helper::ACGT;

void helper::recodeSNPs(CSample & sample)
{

    int nloc = sample.countLocus();
    
    ACGT().resize(nloc);
    
    /////////////////////////
    // Get high / low pairing
    
    for (int i = 1 ; i <= nloc ; i++ ) 
    {

	ACGT(i) = "Z1";

	for (int j = 1 ; j <= sample.countFamily(); j++)
	{
	    for (int k = 1; k <= sample.fam(j)->nInFamily(); k++)
	    {

		if (sample.fam(j)->ind(k)->loc(i).one != "0") 
		{
		    if (sample.fam(j)->ind(k)->loc(i).one < ACGT(i).substr(0,1) ) 
			ACGT(i).replace(0,1,sample.fam(j)->ind(k)->loc(i).one);
		    if (sample.fam(j)->ind(k)->loc(i).one > ACGT(i).substr(1,1) )
			ACGT(i).replace(1,1,sample.fam(j)->ind(k)->loc(i).one);
		}
		
		if (sample.fam(j)->ind(k)->loc(i).two != "0") 
		{
		    if (sample.fam(j)->ind(k)->loc(i).two < ACGT(i).substr(0,1) ) 
			ACGT(i).replace(0,1,sample.fam(j)->ind(k)->loc(i).two);
		    if (sample.fam(j)->ind(k)->loc(i).two > ACGT(i).substr(1,1) )
			ACGT(i).replace(1,1,sample.fam(j)->ind(k)->loc(i).two);
		}

	    }
	}
	
    }
    

    ////////////////////
    // Recode
    
    for (int i = 1 ; i <= nloc ; i++ )
    {
	for (int j = 1 ; j <= sample.countFamily() ; j++)
	{	
	    for (int k = 1 ; k <= sample.fam(j)->nInFamily(); k++)
	    {
		if (sample.fam(j)->ind(k)->loc(i).one == ACGT(i).substr(0,1) ) 
		    sample.fam(j)->ind(k)->loc(i).one = "1";
		if (sample.fam(j)->ind(k)->loc(i).one == ACGT(i).substr(1,1) )
		    sample.fam(j)->ind(k)->loc(i).one = "2";

		if (sample.fam(j)->ind(k)->loc(i).two == ACGT(i).substr(0,1) ) 
		    sample.fam(j)->ind(k)->loc(i).two = "1";
		if (sample.fam(j)->ind(k)->loc(i).two == ACGT(i).substr(1,1) )
		    sample.fam(j)->ind(k)->loc(i).two = "2";
	    }
	}
    }
}

int helper::uSatLocusGivenSNP(Driver &d, int s, BoolArray & est)
{
    if (!par::usat) return s;
    else
    {
	// need to find which locus position 's' sits on
	// s coded 1->H in where H=number of SNPs in haplotype
	// est coded 1->L (T/F) where L=number of LOCI in window
	// 

	int t=1;
	int c=d.win[Window::n]->alt.size();

	// consider each locus specified in the window
	for (int i=0 ; i < c ; i++)
	{

	    // determine the real number of this locus
	    int a = d.win[Window::n]->alt[i] - 1 ;
	    
	    // count over how many alleles / SNP positions this locus takes
	    for (int j=0; j<d.downcode[a].size(); j++)
	    {
		// does this SNP position match queried one?
		// if so, return the LOCUS in est() space
		if (t==s) return i+1;
		
		// move to next SNP position
		t++;
	    }
	}
    }
}


void helper::calcUniqueReducedHaplotypeNum(BoolArray est, 
					   int nhap, 
					   IntArray & betacode, 
					   int & nsubhap,
					   Driver & d)
{

    // a reference to nsubhap passed -- this value will be changed
    nsubhap = 1; 
    // as will this 
    betacode().resize(nhap);

    // Set first element fixed to 1 (first haplotype coefficient is 0)
    betacode(1) = 1; 

    // Assign the remaining haplotypes to a subhaplotype
    for (int i=2 ; i <= nhap ; i++)
    {
	bool match = false;

	for (int j=1; j < i ; j++)
	{
	    // extract sub-haplotype

	    // get this subhap (i)
	    string subhapi = "";
	    for (int ii=1 ; ii <= ind::hap[i].length() ; ii++)
		if (est(helper::uSatLocusGivenSNP(d,ii,est))) 
		    subhapi += ind::hap[i].pos(ii);

	    // get reference subhap (j)
	    string subhapj = "";
	    for (int jj=1 ; jj <= ind::hap[j].length() ; jj++)
		if (est(helper::uSatLocusGivenSNP(d,jj,est))) 
		    subhapj += ind::hap[j].pos(jj);
	    
	    if (subhapi == subhapj ) 
	    {
		betacode(i) = betacode(j);
		match = true;
		break;
	    }
	}
	
	// If no subhaplotype match found 
	if (!match) betacode(i) = ++nsubhap;
	
    }
    

}

void helper::calcUserReducedHaplotypeNum(vector<int> con,
					 int nhap, 
					 IntArray & betacode, 
					 int & nsubhap)
{
    
    // a reference to nsubhap passed -- this value will be changed
    nsubhap = 1; 
    // as will this 
    betacode().resize(nhap);
    
    // Set first element fixed to 1 (first haplotype coefficient is 0)
    if (con[0] != 1) error("Bad --constrain list: 1st element must be 1");
    betacode(1) = 1; 
    
    // e.g. --constrain 1,1,1,1,2,2,3,3/1,1,1,1,2,2,2,2
    // nb. it is up to the user to make sure alt and null are nested...

    // nsubhap is the higest number in --constrain
    for (int i=0; i<con.size(); i++)
    {
	betacode(i+1) = con[i];
	if (con[i] > nsubhap) nsubhap = con[i];	
    }
    
    // ensure the lists match with # observed haps
    if (nsubhap > nhap) error("Bad --constrain list: element greater than # haps.");
    if (con.size() != nhap) error("Bad --constrain list: too many or too few haps.");
}



void helper::extractParameters(Model &m)
{
    
    int cnt = 1;
    if (par::binary)
	if (!par::fix_mean)
	    m.mean = m.estimates[cnt++];
	else
	    m.mean = par::fixed_mean;


    if (!par::binary) 
	if (!par::fix_mean)
	    m.mean = m.estimates[cnt++];
	else
	    m.mean = par::fixed_mean;
    
    m.var = 0;
    if (!par::binary) 
	if (!par::fix_var)
	    m.var = m.estimates[cnt++];
	else
	    m.var = par::fixed_var;

    m.cbeta.setCol(par::covn);
    if (m.cov)
	for (int i=1; i <= par::covn ; i++)
	    m.cbeta[i] = m.estimates[cnt++];

    CMatrix i_tempbeta(par::covn,m.nsubhap);
    m.ibeta.setRow(par::covn);
    m.ibeta.setCol(m.nhap);
    if (m.gxe)
    {
	for (int c=1; c<=par::covn; c++)
	{
	    i_tempbeta[c][1] = 0;
	    for (int h=2; h<=m.nsubhap; h++)
		i_tempbeta[c][h] = m.estimates[cnt++];
	}
    }
    // ...and expand
    for (int c=1; c<=par::covn; c++)
	for (int h=1; h<= m.nhap; h++)
	    m.ibeta[c][h] = i_tempbeta[c][m.betacode(h)];
    

    m.bet_beta.setCol(m.nhap);
    m.bet_tempbeta.setCol(m.nsubhap);

    m.with_beta.setCol(m.nhap);
    m.with_tempbeta.setCol(m.nsubhap);
    
    // extract reduced set of betas
    if (m.total)
    {
	m.bet_tempbeta[1] = m.with_tempbeta[1] = 0;
	for (int i = 2 ; i <= m.nsubhap ; i++) 
	    m.bet_tempbeta[i] = m.with_tempbeta[i] = m.estimates[cnt++];
    }
    else if (m.between)
    {
	m.bet_tempbeta[1] = 0;
	for (int i = 2 ; i <= m.nsubhap ; i++) 
	    m.bet_tempbeta[i] = m.estimates[cnt++];	
    }

    if (m.within && !m.total)
    {
	m.with_tempbeta[1] = 0;
	for (int i = 2 ; i <= m.nsubhap ; i++) 
	    m.with_tempbeta[i] = m.estimates[cnt++];	
    }

    // expand betas, imposing constraints as defined in betacode()
    for (int i = 1 ; i <= m.nhap ; i++) 
    {
	m.bet_beta[i] = m.bet_tempbeta[m.betacode(i)];
	m.with_beta[i] = m.with_tempbeta[m.betacode(i)];
    }

    
    // freqs at end, potentially, but leave here -- they will be in ind::hapfreq1[]
    
}



void helper::sub2fullMapping(Model m,IntArray & s)
{
    s().resize(m.nsubhap);
    for (int i = 1 ; i <= m.nsubhap ; i++)
	for (int j = 1 ; j <= m.nhap ; j++)
	{
	    if (m.betacode(j) == i)
	    {
		s(i) = j;
		break;
	    }
	}
}


void helper::subFrequency(Model & m, CVector & s)
{
    
    for (int i = 1 ; i <= m.nsubhap ; i++) 
	for (int j = 1 ; j <= m.nhap ; j++)
	    if (m.betacode(j) == i) 
	    	s[i] += ind::hapfreq1[j];
}

void helper::sampleVariances(CSample & s)
{

    // initialise, if not already set
    for (int c=1; c <= s.countCovariate(); c++)
    {
	if (!s.cov(c)->mean_set)
	    s.cov(c)->mean = 0;
	if (!s.cov(c)->var_set)
	    s.cov(c)->var = 0;
    }

    for (int q=1; q <= s.countQTrait(); q++)
    {
	if (!s.qt(q)->mean_set)
	    s.qt(q)->mean = 0;
	if (!s.qt(q)->var_set)
	    s.qt(q)->var = 0;
    }

    // SUM
    for (int f=1; f<=s.countFamily(); f++)
    {
	CIndividual * proband = s.fam(f)->proband;
	
	for (int c=1; c <= s.countCovariate(); c++)
	    if (!s.cov(c)->mean_set)
		s.cov(c)->mean += proband->cov(c);

	for (int q=1; q <= s.countQTrait(); q++)
	    if (!s.qt(q)->mean_set)
	    s.qt(q)->mean += proband->qt(q);	
    }

    // SUM / N
    for (int c=1; c <= s.countCovariate(); c++)
	if (!s.cov(c)->mean_set)
	    s.cov(c)->mean /= s.countFamily();
    
    for (int q=1; q <= s.countQTrait(); q++)
	if (!s.qt(q)->mean_set)
	    s.qt(q)->mean /= s.countFamily();
    
    // SUM OF SQUARES
    for (int f=1; f<=s.countFamily(); f++)
    {
	CIndividual * proband = s.fam(f)->proband;
	
	for (int c=1; c <= s.countCovariate(); c++)
	    if (!s.cov(c)->var_set)
		s.cov(c)->var += mymath::sqr(proband->cov(c) - s.cov(c)->mean);
	
	for (int q=1; q <= s.countQTrait(); q++)
	    if (!s.qt(q)->var_set)
		s.qt(q)->var += mymath::sqr(proband->qt(q) - s.qt(q)->mean);
    }

    // SUM OF SQUARES / N-1
    for (int c=1; c <= s.countCovariate(); c++)
	if (!s.cov(c)->var_set)
	    s.cov(c)->var /= (s.countFamily()-1);
    for (int q=1; q <= s.countQTrait(); q++)
	if (!s.qt(q)->var_set)
	    s.qt(q)->var /= (s.countFamily()-1);
    

}



void helper::summary(CSample &s)
{
    
    if (par::nice_dump) return;
    
    int n_trio = 0;
    int n_unrel = 0;
    
    for (int f=1; f<=s.countFamily(); f++)
    {
	if (s.fam(f)->proband->box->parents) n_trio++;
	else n_unrel++;
    }
    
    cout << n_unrel << " individuals w/out parents. " 
	 << n_trio << " individuals with parents. ";
    if (par::binary) 
    { 
	cout << "Binary trait: ";
	if (par::fix_mean) cout << "Prevalence fixed to " << 1/(1+exp(-par::fixed_mean)) << ". ";
    }
    if (!par::binary)
    {
	cout << "Continuous trait: ";
	if (par::fix_mean) cout << "Mean fixed to " << par::fixed_mean << ". ";
	if (par::fix_var) cout << "Variance fixed to " << par::fixed_var << ". ";
    }
    cout << "\n";
}


void helper::collapsePhases(CVector &K, CMatrix &B, CMatrix &W)
{

    // use temporary hash_map to collapse phases -- should be faster
    map<helper::vpair, double> bw_hash;
    
    // take each row of B/K/W
    for (int i=1;i<=B.getRow();i++)
    {
	helper::vpair vp;
	vp.B = B[i];
	vp.W = W[i];
	bw_hash[vp] += K[i];    
    }

    // copy from bw_hash back to B and W 
    K.setCol(bw_hash.size());
    B.setRow(bw_hash.size());
    W.setRow(bw_hash.size());

    int j=1;
    map<helper::vpair,double>::iterator i;
    for (i=bw_hash.begin(); i!=bw_hash.end(); i++)
    {
	helper::vpair vp = i->first;
	K[j] = i->second;
	B[j] = vp.B;
	W[j] = vp.W;
	j++;
    }
    
}



void helper::usatFile(string f, 
		      vector<int>  & acount, 
		      vector< vector<int> > & downcode,
		      int l)
{

    acount.resize(0);
    downcode.resize(0);

    ifstream I;
    I.open(f.c_str());
    while(!I.eof())
    {
	int t;
	I >> t;
	acount.push_back(t);
    }
    I.close();

    // remove last item...
    acount.erase(acount.end()-1);

    int s=0;
    for (int i=0;i<acount.size();i++)
	s+=acount[i];

    if (s!=l) error("Something wrong with the .all file...");
    
    downcode.resize(acount.size());
    s=0;
    for (int i=0;i<downcode.size();i++)
	for (int j=0; j < acount[i]; j++)
	    downcode[i].push_back(++s);
    
}


void helper::niceDumper(Driver & d, CSample & sample)
{
    // consider each family
    for (int f=1; f<=sample.countFamily(); f++)
    {
	
	CIndividual * proband = sample.fam(f)->proband; 
	
	// If an individual...
	if (!proband->box->parents)
	{
	    // number of phases for this individual 
	    int n = proband->box->K.getCol();
	    
	    // perform our own collapse function on the fly
	    for (int p = 0 ; p < n ; p++)
	    {
		for (int p2 = 0 ; p2 < n ; p2++)
		{
		    // a matching phase?
		    
		    if ( ( proband->box->P[p].paternal == proband->box->P[p2].paternal 
			   && proband->box->P[p].maternal == proband->box->P[p2].maternal )
			 ||
			 ( proband->box->P[p].paternal == proband->box->P[p2].maternal 
			   && proband->box->P[p].maternal == proband->box->P[p2].paternal
			     ) )
		    {
			if (proband->box->K[p+1] >= 0 && proband->box->K[p2+1] >= 0 && p!=p2)
			{
			    proband->box->K[p+1] += proband->box->K[p2+1];
			    proband->box->K[p2+1] = -1;
			}
		    }
		}
	    }
	    
	    // perform our own collapse function on the fly
	    int c=0;
	    for (int p = 0 ; p < n ; p++)
	    {
		if (proband->box->K[p+1] >= 0)
		{
		    c++;
		    cout << proband->family->getID()
			 << "\t"
			 << proband->getID()
			 << "\t"
			 << c
			 << "\t"
			 << helper::getHapName(d,proband->box->P[p].paternal)
			 << " "
			 << helper::getHapName(d,proband->box->P[p].maternal)
			 << "   " << proband->box->K[p+1]
			 << "\n";
		}
	    }
	}
	else // ... otherwise consider as a trio
	{

	    // CHECK: standardize K so it sums to 1 
	    // CHECK: doesn't effect the test is does not sum to 1...
	    
	    double sumK = proband->box->K.sum();
	    if (sumK>0)
		proband->box->K.multiply(1/sumK);
	    
	    // number of PPhases/Phases for this individual 
	    int n = proband->box->K.getCol();
	    
	    // perform our own collapse function on the fly
	    int cnt = 0;
	    int cnt2 = 0;

	    for (int pp=0; pp < proband->box->PP.size(); pp++)
		for (int p=0; p < proband->box->PP[pp].P.size(); p++)
		{
		    cnt++;
		    cnt2 = 0 ;

		    for (int pp2=0; pp2 < proband->box->PP.size(); pp2++)
			for (int p2=0; p2 < proband->box->PP[pp2].P.size(); p2++)
			{
			    cnt2++;
			    
			    // a matching phase (have to collapse PP as well...) ?
			    
			    // individual match?
			    if ( (    proband->box->PP[pp].P[p].paternal == proband->box->PP[pp2].P[p2].paternal 
				      && proband->box->PP[pp].P[p].maternal == proband->box->PP[pp2].P[p2].maternal )
				 ||
				 (    proband->box->PP[pp].P[p].paternal == proband->box->PP[pp2].P[p2].maternal 
				      && proband->box->PP[pp].P[p].maternal == proband->box->PP[pp2].P[p2].paternal
				     ) )
			    {
				
				// father match match?
				if ( (    proband->box->PP[pp].father.paternal == proband->box->PP[pp2].father.paternal 
					  && proband->box->PP[pp].father.maternal == proband->box->PP[pp2].father.maternal )
				     ||
				     (    proband->box->PP[pp].father.paternal == proband->box->PP[pp2].father.maternal 
					  && proband->box->PP[pp].father.maternal == proband->box->PP[pp2].father.paternal
					 ) )
				{
				    
				    // mother match?
				    if ( (    proband->box->PP[pp].mother.paternal == proband->box->PP[pp2].mother.paternal 
					      && proband->box->PP[pp].mother.maternal == proband->box->PP[pp2].mother.maternal )
					 ||
					 (    proband->box->PP[pp].mother.paternal == proband->box->PP[pp2].mother.maternal 
					      && proband->box->PP[pp].mother.maternal == proband->box->PP[pp2].mother.paternal
					     ) )
				    {
					


					if (proband->box->K[cnt] >= 0 && proband->box->K[cnt2] >= 0 && cnt!=cnt2)
					{
					    proband->box->K[cnt] += proband->box->K[cnt2];
					    proband->box->K[cnt2] = -1;
					}
				    }
				}
			    }
			}   
		}
	    
	    // perform our own collapse function on the fly
	    int c=0;
	    cnt = 0;
	    for (int pp=0; pp < proband->box->PP.size(); pp++)
		for (int p=0; p < proband->box->PP[pp].P.size(); p++)
		{
		    cnt++;
		    
		    if (proband->box->K[cnt] >= 0)
		    {
			c++;
			cout << proband->family->getID()
			     << "\t"
			     << proband->getID()
			     << "\t"
			     << c
			     << "\t"
			     << helper::getHapName(d,proband->box->PP[pp].P[p].paternal)
			     << " "
			     << helper::getHapName(d,proband->box->PP[pp].P[p].maternal)
			     << "   " << proband->box->K[cnt]
			     << "   [ " 
			     << helper::getHapName(d,proband->box->PP[pp].father.paternal)
			     << " " 
			     << helper::getHapName(d,proband->box->PP[pp].father.maternal)
			     << " x " 
			     << helper::getHapName(d,proband->box->PP[pp].mother.paternal)
			     << " " 
			     << helper::getHapName(d,proband->box->PP[pp].mother.maternal)
			     << " ] " 
			     << "\n";

		    }
		}
	}
	    	
    }

}





string helper::getHapName(Driver &d, CHaplotype &h)
{

    string hapname;
    if (!par::usat)
	for (int l=1; l <= h.pos().size(); l++)
	    hapname += helper::ACGT(d.win[Window::n]->alt[l-1]).substr(atoi( h.pos(l).c_str())  -1,1);
    else
    {
	//
	// need to consider usat upcoding
	//

	// initialise position in haplotype
	int l2 = 1;
	
	// each locus specified in --alt
	for (int l=0; l < d.win[Window::n]->alt.size(); l++)
	{
	    // store locus number
	    int a = d.win[Window::n]->alt[l] - 1 ;
	    
	    // calculate position in haplotype
	    
	    if (d.acount[a]==1) // a SNP
	    {
		int s = d.downcode[a][0];
		hapname += helper::ACGT(s).substr(atoi( h.pos(l2).c_str())-1,1);
		l2++;
	    }
	    else // or a uSAT
	    {
 		for (int k=0; k<d.acount[a]; k++)
		{
		    if (h.pos(l2)=="2")
 		    {
 			string s=d.sample.loc(d.downcode[a][k])->name;
 			hapname += s.substr(s.find_last_of("["));
 		    }
		    l2++;
		}
	    }
	}
    }
    return hapname+" ";
}


double helper::combinedEmpiricalPValue(vector<double>& d1, vector<double>& d2)
{
    double o1 = 0;
    double o2 = 0;
    
    for (int i=0; i < d1.size(); i++)
    {	
	if (d1[i] >= d1[0]) o1++; 
	if (d2[i] >= d2[0]) o2++;
    }

    o1 /= d1.size();
    o2 /= d2.size();
    
    double oc = -2*(log(o1)+log(o2)) ;
    
    vector<double> pc;

    // Calculate for other perms
    
    for (int i=1; i<d1.size(); i++)
    {    
	double p1 = 0;
	double p2 = 0;
	
	for (int j=1; j<d1.size(); j++)
	{
	    if (d1[j] >= d1[i]) p1++; 
	    if (d2[j] >= d2[i]) p2++; 	    
	}
	
	p1 /= d1.size()-1;
	p2 /= d1.size()-1;

	pc.push_back(-2*(log(p1)+log(p2)));
    }
    
    double c = 0;
    for (int i=0; i < pc.size(); i++)
        if (oc <= pc[i]) c++;
    
    c++;
    c /= pc.size()+1;

    return c;
}
