#include "driver.h"
#include "haplotype.h"
#include "permutation.h"
#include "parameters.h"

void Driver::processWindow()
{
    cout << "\nResults of sliding window analysis\n";

    // Primary LRT 
    // Secondary test (global only)


    /////////////////////////////////////////
    // Calculate primary positional score
    
    // Number of instances a position appears in a window
    IntArray pos_primary_count;
    pos_primary_count().resize(total_nloci);
    pos_primary_count().clear();

    // global secondary test 
    IntArray pos_secondary_count;
    pos_secondary_count().resize(total_nloci);
    pos_secondary_count().clear();

    // largest local secondary test 
    IntArray pos_combined_count;
    pos_combined_count().resize(total_nloci);
    pos_combined_count().clear();



    // For original, and permutations, summed scores
    CVector pos_primary(total_nloci);
    CMatrix pos_primary_perm(par::perm,total_nloci);
    pos_primary.clear();
    pos_primary_perm.clear();

    CVector pos_secondary(total_nloci);
    CMatrix pos_secondary_perm(par::perm,total_nloci);
    pos_secondary.clear();
    pos_secondary_perm.clear();

    CVector pos_combined(total_nloci);
    CMatrix pos_combined_perm(par::perm,total_nloci);
    pos_combined.clear();
    pos_combined_perm.clear();

    // primary score will have a valid value for every window
    // secondary score may not -- if only 2 unique haplotypes, for example.
    
    for (int i=0; i<=Window::n; i++)
    {
	for (int j=1; j<= total_nloci; j++)
	{

	    // does this window test this position?
	    // i.e. in --alt but not in --null? 
	    bool in_alt = false;
	    bool in_null = false;

	    for (int k=0; k < win[i]->alt.size() ; k++)
		if (win[i]->alt[k] == j) in_alt = true;

	    for (int k=0; k < win[i]->null.size() ; k++)
		if (win[i]->null[k] == j) in_null = true;

	    if (in_alt && !in_null)
	    {
		// Original primary score (LRT)
		if (win[i]->results.primary[0] != -1)
		{
		    pos_primary[j] += win[i]->results.primary[0];
		}
		
		// Only secondary global or LL test, and valid result (!=-999)
		if (par::secondary)
		{
		    // global
		    if (par::sec_global)
		    {
			if (win[i]->results.secondary[0][0] != -999)
			    pos_secondary[j] += win[i]->results.secondary[0][0];
		    }
		    else // lagrest local
		    {		
			if (win[i]->results.secondary[0][1] != -999)
			    pos_secondary[j] += win[i]->results.secondary[0][1];
		    }
		    }


		//////////////////
		// Permuted scores

		for (int p=1; p <= par::perm; p++)
		{
		    // Primary test
		    if (win[i]->results.primary[0] != -1)
		    {
			pos_primary_perm[p][j] += win[i]->results.primary[p];
		    }

		    // Secondary test
		    if (par::secondary)
		    {
			if (win[i]->results.secondary[p][0] != -999)
			    pos_secondary_perm[p][j] += win[i]->results.secondary[p][0];
			if (win[i]->results.secondary[p][1] != -999)
			    pos_combined_perm[p][j] += win[i]->results.secondary[p][1];
		    }
		} 
		
                // Count number of times marker included
		if (win[i]->results.primary[0] != -1)
		    pos_primary_count(j)++;
		
		if (par::secondary)
		{
		    if (win[i]->results.secondary[0][0] != -999)
			pos_secondary_count(j)++;
		    if (win[i]->results.secondary[0][1] != -999)
			pos_combined_count(j)++;
		}		
	    }
	}
    }

    ////////////////////////////////////////////////////////////
    // Adjust score for number of instances appears in a window

    for (int j=1; j<= total_nloci; j++)
    {
	// Primary scores
	if (pos_primary_count(j) > 0 )
	    pos_primary[j] /= (double)pos_primary_count(j);
	else
	    pos_primary[j] = -1;
	
	// Secondary scores
	if (par::secondary)
	{
	    if (pos_secondary_count(j) > 0)
		pos_secondary[j] /= (double)pos_secondary_count(j);
	    else
		pos_secondary[j] = -999;
	
	    if (pos_combined_count(j) > 0)
		pos_combined[j] /= (double)pos_combined_count(j);
	    else
		pos_combined[j] = -999;
	}


	// Permuted scores
	for (int p=1; p<= par::perm; p++)
	{
	    // Primary
	    if (pos_primary_count(j) > 0 )
		pos_primary_perm[p][j] /= (double)pos_primary_count(j);
	    else
		pos_primary_perm[p][j] = -1;

	    // Secondary
	    if (par::secondary)
	    {
		if (pos_secondary_count(j) > 0 )
		    pos_secondary_perm[p][j] /= (double)pos_secondary_count(j);
		else
		    pos_secondary_perm[p][j] = -999;

		if (pos_combined_count(j) > 0 )
		    pos_combined_perm[p][j] /= (double)pos_combined_count(j);
		else
		    pos_combined_perm[p][j] = -999;
		
	    }
	}
    }

    
    ///////////////////////////////////////////
    // Calculate maximum, minumum, total area

    double maximum_primary = -99;
    double sum_primary = 0;

    double maximum_secondary = -99;
    double sum_secondary = 0;

    double maximum_combined = -99;
    double sum_combined = 0;

    for (int j=1; j <= total_nloci; j++)
    {
	// maximum
	if (pos_primary[j] > maximum_primary)
	    maximum_primary = pos_primary[j];

	if (par::secondary 
	    && pos_secondary[j] != -999
	    && pos_combined[j] != -999)
	{
	    if (mymath::sqr(pos_secondary[j]) > maximum_secondary)
		maximum_secondary = mymath::sqr(pos_secondary[j]);

	    if (mymath::sqr(pos_combined[j]) > maximum_combined)
		maximum_combined = mymath::sqr(pos_combined[j]);
	    
	}
    }
    
    // Calculate Sum scores
    for (int j=1; j <= total_nloci; j++)
    {
	if (pos_primary[j] != -1)
	    sum_primary += pos_primary[j];
	
	if (par::secondary && pos_secondary[j] != -999)
	    sum_secondary += mymath::sqr(pos_secondary[j]);

	if (par::secondary && pos_combined[j] != -999)
	    sum_combined += mymath::sqr(pos_combined[j]);
    }


    // Calculate p-values for each position, maximum, total area
    // Also calculate average pos_primary_perm[][]

    CVector pos_primary_perm_p(total_nloci);
    pos_primary_perm_p.clear();

    CVector pos_secondary_perm_p(total_nloci);
    pos_secondary_perm_p.clear();

    CVector pos_combined_perm_p(total_nloci);
    pos_combined_perm_p.clear();
    
    double maximum_primary_perm_p = 0;
    double sum_primary_perm_p = 0;
    double maximum_secondary_perm_p = 0;
    double sum_secondary_perm_p = 0;
    double maximum_combined_perm_p = 0;
    double sum_combined_perm_p = 0;

    // Average permutation values
    CVector pos_primary_perm_avr(total_nloci);
    pos_primary_perm_avr.clear();
    
    CVector pos_secondary_perm_avr(total_nloci);
    pos_secondary_perm_avr.clear();
    
    CVector pos_combined_perm_avr(total_nloci);
    pos_combined_perm_avr.clear();

    for (int p=1; p <= par::perm; p++)
    {
	
       	// Calculate maximum of this permutation
	// Calculate sum of this permutation
	double this_primary_perm_maximum = -99 ;
	double this_primary_perm_sum = 0;

	double this_secondary_perm_maximum = -99 ;
	double this_secondary_perm_sum = 0;

	double this_combined_perm_maximum = -99 ;
	double this_combined_perm_sum = 0;

	for (int j=1; j <= total_nloci; j++)
	{
	    // Maximum
	    if (pos_primary_perm[p][j] > this_primary_perm_maximum) 
		this_primary_perm_maximum = pos_primary_perm[p][j];
	    
	    if (par::secondary 
		&& pos_secondary_perm[p][j] != -999
		&& pos_combined_perm[p][j] != -999)
	    {
		if (mymath::sqr(pos_secondary_perm[p][j]) > this_secondary_perm_maximum) 
		    this_secondary_perm_maximum = mymath::sqr(pos_secondary_perm[p][j]);
		
		if (mymath::sqr(pos_combined_perm[p][j]) > this_combined_perm_maximum) 
		    this_combined_perm_maximum = mymath::sqr(pos_combined_perm[p][j]);
	    }
		
	}
	
	// Sum
	for (int j=1; j <= total_nloci; j++)
	{
	    if (pos_primary_perm[p][j] != -1)
		this_primary_perm_sum += pos_primary_perm[p][j];
	    
	    if (par::secondary && pos_secondary_perm[p][j] != -999)
		this_secondary_perm_sum += mymath::sqr(pos_secondary_perm[p][j]);
	    
	    if (par::secondary && pos_combined_perm[p][j] != -999)
		this_combined_perm_sum += mymath::sqr(pos_combined_perm[p][j]);
	    
	}

	// Update average under permuted nulls
	for (int j=1; j <= total_nloci; j++)
	{
	    if (pos_primary_perm[p][j] != -1)
		pos_primary_perm_avr[j] += pos_primary_perm[p][j];
	    
//	    if (par::secondary && pos_secondary_perm[p][j] != -999)
//		pos_secondary_perm_avr[j] += pos_secondary_perm[p][j];

//	    if (par::secondary && pos_combined_perm[p][j] != -999)
//		pos_combined_perm_avr[j] += pos_combined_perm[p][j];

	}


	// Count times exceeded, primary and secondary

	// Max
	if (this_primary_perm_maximum > maximum_primary)
	    maximum_primary_perm_p++;

	if (par::secondary)
	{
	    if (this_secondary_perm_maximum > maximum_secondary)
		maximum_secondary_perm_p++;
	    if (this_combined_perm_maximum > maximum_combined)
		maximum_combined_perm_p++;
	}

	// Sum
	if (this_primary_perm_sum > sum_primary)
	    sum_primary_perm_p++;
	
	if (par::secondary)
	{
	    if (this_secondary_perm_sum > sum_secondary)
		sum_secondary_perm_p++;
	    if (this_combined_perm_sum > sum_combined)
		sum_combined_perm_p++;
	}
	
	
        // Position-specific
	for (int j=1; j <= total_nloci; j++)
	    if (pos_primary_perm[p][j] > pos_primary[j])
		pos_primary_perm_p[j]++;

	if (par::secondary)
	{
	    for (int j=1; j <= total_nloci; j++)
		if (pos_secondary_perm[p][j] != -999)
		    if (mymath::sqr(pos_secondary_perm[p][j]) > mymath::sqr(pos_secondary[j]))
			pos_secondary_perm_p[j]++;
	
	    for (int j=1; j <= total_nloci; j++)
		if (pos_combined_perm[p][j] != -999)
		    if (mymath::sqr(pos_combined_perm[p][j]) > mymath::sqr(pos_combined[j]))
			pos_combined_perm_p[j]++;
	    
	}
	
    }

    // Calculate empirical p-values
    maximum_primary_perm_p =  (maximum_primary_perm_p+1.0) / (double)(par::perm+1);
    sum_primary_perm_p = (sum_primary_perm_p+1.0) / (double)(par::perm+1);
    for (int j=1; j <= total_nloci; j++)
	pos_primary_perm_p[j] = (pos_primary_perm_p[j]+1.0)  /  (double)(par::perm+1);
    
    if (par::secondary)
    {
	maximum_secondary_perm_p =  (maximum_secondary_perm_p+1.0) / (double)(par::perm+1);
	sum_secondary_perm_p = (sum_secondary_perm_p+1.0) / (double)(par::perm+1);
	for (int j=1; j <= total_nloci; j++)
	    pos_secondary_perm_p[j] = (pos_secondary_perm_p[j]+1.0)  /  (double)(par::perm+1);

	maximum_combined_perm_p =  (maximum_combined_perm_p+1.0) / (double)(par::perm+1);
	sum_combined_perm_p = (sum_combined_perm_p+1.0) / (double)(par::perm+1);
	for (int j=1; j <= total_nloci; j++)
	    pos_combined_perm_p[j] = (pos_combined_perm_p[j]+1.0)  /  (double)(par::perm+1);
	
    }
    
    // Calculate average under null
    for (int j=1; j <= total_nloci; j++)
	pos_primary_perm_avr[j] /= (double)(par::perm);

//    if (par::secondary)
    //   {
//	for (int j=1; j <= total_nloci; j++)
//	    pos_secondary_perm_avr[j] /= (double)(par::perm);
    //	for (int j=1; j <= total_nloci; j++)
//	    pos_combined_perm_avr[j] /= (double)(par::perm);
//    }

    /////////////////////////////////////////////////
    // Display results
    
    // Optinally display raw sliding window scores
    if (par::win_all_results)
    {
	cout << "\nBasic sliding window scores\n";
	cout << "---------------------------\n";
	for (int i=0; i<=Window::n; i++)
	{
	    cout << "\nWindow starting position " << i+1;
	    cout << "\t" << win[i]->results.primary[0];
	    if (par::secondary) 
		cout << "\t" << win[i]->results.secondary[0][0]
		     << "\t" << win[i]->results.secondary[0][1];
	}
    }
    
    // Summary permutation tests
    cout << "\nGlobal permutation tests";
    cout << "\n------------------------\n";    
    cout << "P_MAX = " << maximum_primary << "\t";
    cout.precision(10);
    cout << " p = " << maximum_primary_perm_p; 
    cout.precision(3);
    cout << "\n";
    cout << "P_SUM = " << sum_primary << "\t";
    cout.precision(10);
    cout << " p = " << sum_primary_perm_p ;
    cout.precision(3);
    cout << "\n";    
    
    if (par::secondary)
    {
	cout << "S_MAX = " << maximum_secondary << "\t" ;
	cout.precision(10);
	cout << " p = " << maximum_secondary_perm_p ;
	cout.precision(3);
	cout << "\n";
	cout << "S_SUM = " << sum_secondary << "\t";
	cout.precision(10);
	cout << " p = " << sum_secondary_perm_p; 
	cout.precision(3);
	cout << "\n";    

// Possibly re-introduce this code...
//	cout << "C_MAX = " << maximum_combined << "\t" ;
//	cout.precision(10);
//	cout << " p = " << maximum_combined_perm_p ;
//	cout.precision(3);
//	cout << "\n";
//	cout << "C_SUM = " << sum_combined << "\t";
//	cout.precision(10);
//	cout << " p = " << sum_combined_perm_p; 
//	cout.precision(3);
//	cout << "\n";    

    }
    
    cout << "\nLocal permutation tests";
    cout << "\n-----------------------\n";    

    for (int j=1; j <= total_nloci; j++)
    {
	// find locus name

	string locname;
	if (!par::usat) locname = sample.loc(j)->name;
	else {
	    locname=sample.loc(downcode[j-1][0])->name;
	    locname = locname.substr(0,locname.find_last_of("_"));
	}
	
	if (!par::usat) 
	  cout << ">> " << locname << " "  << locmap.bp[sample.loc(j)->name] << " ";
	else
	  cout << ">> " << locname << " "  << locmap.bp[sample.loc(downcode[j-1][0])->name] << " ";

	cout << "P_" << j 
	     << "= " << pos_primary[j] << "\t";
	
	if (pos_primary[j] != -1)
	{
	    cout.precision(10);
	    cout << " p= " << pos_primary_perm_p[j] << "\t";
	    cout.precision(3);
	}
	else
	    cout << " p= nan\t";
	
//	cout << " D= " << pos_primary[j] - pos_primary_perm_avr[j] << "\t\t";
	
	if (par::secondary)
	{
	    
            // Global secondary test
	    cout << " S_" << j 
		 << "= " << pos_secondary[j] << "\t";

	    if (pos_secondary[j] != -999)
	    {
		cout.precision(10);
		cout << " p= " << pos_secondary_perm_p[j] << "\t";
		cout.precision(3);
	    }
	    else
		cout << " p= nan\t";
	    
//	    cout<< " Dg= " << pos_secondary[j] - pos_secondary_perm_avr[j];


            // Combined pri+sec p-values (empirically combined) 
// TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE 
// TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE 
// TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE 
//	    cout << " C_" << j 
//		 << "= " << pos_combined[j] << "\t";
//
//	    if (pos_combined[j] != -999)
//	    {
//		cout.precision(10);
//		cout << " p= " << pos_combined_perm_p[j] << "\t";
//		cout.precision(3);
//	    }
//	    else
//		cout << " p= nan\t";
// TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE 
// TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE 
// TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE TMP REMOVE 

	    
//	    cout<< " Dll= " << pos_combined[j] - pos_combined_perm_avr[j];

	}
	cout << "\n";
	
    }
    
    
    // Optinally, output all permuted positional scores
    
    if (par::win_all_results)
    {
	
	// Primary
	for (int j=1; j<= total_nloci; j++)
	{
	    cout << "\n" << j << "\t" << pos_primary[j];
	    for (int p=1; p<= par::perm; p++)
		cout << "\t" << pos_primary_perm[p][j];
	}	 
	cout << "\n\n";
	
	// Secondary
	if (par::secondary)
	{
	    for (int j=1; j<= total_nloci; j++)
	    {
		cout << "\n" << j << "\t" << pos_secondary[j];
		for (int p=1; p<= par::perm; p++)
		    cout << "\t" << pos_secondary_perm[p][j];
	    }	 

	    cout << "\n";

	    for (int j=1; j<= total_nloci; j++)
	    {
		cout << "\n" << j << "\t" << pos_combined[j];
		for (int p=1; p<= par::perm; p++)
		    cout << "\t" << pos_combined_perm[p][j];
	    }	 

	}
	cout << "\n";
    }	
    
}


void Driver::storeOriginal()
{
    // ASSUMPTION: one proband per family

    int n = sample.countFamily(); 

    if (par::binary) 
    {
	original_bt.resize(n);
	for (int i=1; i <= sample.countFamily(); i++)
	    original_bt[i-1] = sample.fam(i)->proband->bt(1);
    }
    else 
    {
	original_qt.resize(n);
	for (int i=1; i <= sample.countFamily(); i++)
	    original_qt[i-1] = sample.fam(i)->proband->qt(1);
    }

}


void Driver::prePermute()
{
    
    if (par::debug)
	cout << "\nabout to pre-permute()\n";

    // ASSUMPTION: TRIOS or UNRELATEDS
    // NOTE: permuting within sibships will be more involved
    int n = sample.countFamily(); 
    
    // Generate and store matrix of permuted scores 
    // for sliding window analysis
    preperm.setRow(par::perm);
    preperm.setCol(n);

    for (int p=1; p <= par::perm ; p++)
    {
	// Create permutation list
	vector<int> pcode(n);
	mymath::permute(pcode);

	vector<double> trait;
	trait.resize(n);
	for (int i=1; i <= sample.countFamily(); i++)
	    if (par::binary)
		trait[i-1] = original_bt[i-1];
	    else 
		trait[i-1] = original_qt[i-1];
	
        // Copy permuted version into sample[]
	for (int i=0; i < n ; i++)
	    preperm[p][i+1] = trait[pcode[i]];
    }

    if (par::debug)
	cout << "\nfinished pre-permute()\n";


}



void Driver::prePermuteW()
{
    // Generate and store matrix of 1s and -1s
    // for sliding window analysis
    
    int n = sample.countFamily(); 
    wpreperm.setRow(par::perm);
    wpreperm.setCol(n);
    
    for (int p=1; p <= par::perm ; p++)
	for (int i=1; i <= n ; i++)
	    if (CRandom::rand() < 0.5)
		wpreperm[p][i] = 1;
	    else
		wpreperm[p][i] = -1;
}



void Driver::permuteW(CMatrix &W)
{
    // for trait-ascertained data, e.g. all affected TDT samples, 
    // we can permute the W matrix instead of permuting the trait
    // with equal probability, replace W with either W or -W
    
    if (CRandom::rand() < 0.5)
	for (int r=1; r<=W.getRow(); r++)
	    for (int c=1; c<=W.getCol(); c++)
		W[r][c] = -W[r][c];
}

